@@ -15,6 +15,13 @@ namespace MiniZinc {
   Expression::equal(const Expression* e0, const Expression* e1) {
     if (e0==e1) return true;
     if (e0 == NULL || e1 == NULL) return false;
+    if (e0->isUnboxedInt() || e1->isUnboxedInt()) return false;
+    if (e0->isUnboxedFloatVal() || e1->isUnboxedFloatVal()) {
+      if (e0->isUnboxedFloatVal() && e1->isUnboxedFloatVal()) {
+        return e0->unboxedFloatToFloatVal()==e1->unboxedFloatToFloatVal();
+      }
+      return false;
+    }
     if (e0->_id != e1->_id) return false;
     if (e0->type() != e1->type()) return false;
     if (e0->hash() != e1->hash()) return false;
@@ -23,6 +30,11 @@ namespace MiniZinc {
 
   inline void
   Expression::type(const Type& t) {
+    if (isUnboxedVal()) {
+      assert(!isUnboxedInt() || t == Type::parint());
+      assert(!isUnboxedFloatVal() || t == Type::parfloat());
+      return;
+    }
     if (eid()==E_VARDECL) {
       this->cast<VarDecl>()->id()->_type = t;
     } else if (eid()==E_ID && this->cast<Id>()->decl()) {
@@ -40,9 +52,117 @@ namespace MiniZinc {
 
   inline IntLit*
   IntLit::a(MiniZinc::IntVal v) {
-    if (v >= -constants().maxConstInt && v <= constants().maxConstInt)
-      return constants().integers->v()[v.toInt()+constants().maxConstInt]->cast<IntLit>();
-    return new IntLit(Location().introduce(), v);
+    
+    if (v.isFinite()) {
+      IntLit* ret = intToUnboxedInt(v.toInt());
+      if (ret) {
+        return ret;
+      }
+    }
+    
+    std::unordered_map<IntVal, WeakRef>::iterator it = constants().integerMap.find(v);
+    if (it==constants().integerMap.end() || it->second()==NULL) {
+      IntLit* il = new IntLit(Location().introduce(), v);
+      if (it==constants().integerMap.end()) {
+        constants().integerMap.insert(std::make_pair(v, il));
+      } else {
+        it->second = il;
+      }
+      return il;
+    } else {
+      return it->second()->cast<IntLit>();
+    }
+  }
+  
+  inline IntLit*
+  IntLit::aEnum(IntVal v, unsigned int enumId) {
+    if (enumId==0)
+      return a(v);
+    IntLit* il = new IntLit(Location().introduce(), v);
+    Type tt(il->type());
+    tt.enumId(enumId);
+    il->type(tt);
+    return il;
+  }
+  
+  inline ASTString
+  Location::LocVec::filename(void) const {
+    return static_cast<ASTStringO*>(_data[0]);
+  }
+  inline unsigned int
+  Location::LocVec::first_line(void) const {
+    if (_size==2) {
+      static const unsigned int pointerBits = sizeof(IntLit*)*8;
+      IntLit* il = static_cast<IntLit*>(_data[1]);
+      long long unsigned int mask = pointerBits<=32 ? 0xFF : 0xFFFFF;
+      union {
+        long long int i;
+        unsigned long long int u;
+      } ui;
+      ui.i = il->v().toInt();
+      return static_cast<unsigned int>(ui.u & mask);
+    } else {
+      IntLit* il = static_cast<IntLit*>(_data[1]);
+      return il->v().toInt();
+    }
+  }
+  inline unsigned int
+  Location::LocVec::last_line(void) const {
+    if (_size==2) {
+      static const unsigned int pointerBits = sizeof(IntLit*)*8;
+      IntLit* il = static_cast<IntLit*>(_data[1]);
+      long long unsigned int first_line_size = pointerBits<=32 ? 8 : 20;
+      long long unsigned int mask = pointerBits<=32 ? 0xFF : 0xFFFFF;
+      long long unsigned int offsetmask = pointerBits<=32 ? 0x7F : 0xFFFFF;
+      union {
+        long long int i;
+        unsigned long long int u;
+      } ui;
+      ui.i = il->v().toInt();
+      // return first line (8 bit) + offset (7 bit)
+      return static_cast<unsigned int>( (ui.u & mask) + ((ui.u >> first_line_size) & offsetmask));
+    } else {
+      IntLit* il = static_cast<IntLit*>(_data[2]);
+      return il->v().toInt();
+    }
+  }
+  inline unsigned int
+  Location::LocVec::first_column(void) const {
+    if (_size==2) {
+      static const unsigned int pointerBits = sizeof(IntLit*)*8;
+      IntLit* il = static_cast<IntLit*>(_data[1]);
+      long long unsigned int first_col_offset = pointerBits<=32 ? 8+7 : 20+20;
+      long long unsigned int mask = pointerBits<=32 ? 0x3F : 0x3FF;
+      union {
+        long long int i;
+        unsigned long long int u;
+      } ui;
+      ui.i = il->v().toInt();
+      // return first line (8 bit) + offset (7 bit)
+      return static_cast<unsigned int>( (ui.u >> first_col_offset) & mask);
+    } else {
+      IntLit* il = static_cast<IntLit*>(_data[3]);
+      return il->v().toInt();
+    }
+  }
+  inline unsigned int
+  Location::LocVec::last_column(void) const {
+    if (_size==2) {
+      static const unsigned int pointerBits = sizeof(IntLit*)*8;
+      IntLit* il = static_cast<IntLit*>(_data[1]);
+      long long unsigned int last_col_offset = pointerBits<=32 ? 8+7+6 : 20+20+10;
+      long long unsigned int mask = pointerBits<=32 ? 0x7F : 0x3FF;
+      union {
+        long long int i;
+        unsigned long long int u;
+      } ui;
+      ui.i = il->v().toInt();
+      // return first line (8 bit) + offset (7 bit)
+      return static_cast<unsigned int>( (ui.u >> last_col_offset) & mask);
+    } else {
+      IntLit* il = static_cast<IntLit*>(_data[4]);
+      return il->v().toInt();
+    }
   }
   
   inline
@@ -51,23 +171,57 @@ namespace MiniZinc {
     rehash();
   }
 
+  inline FloatLit*
+  FloatLit::a(MiniZinc::FloatVal v) {
+    if (sizeof(double) <= sizeof(FloatLit*) && v.isFinite()) {
+      FloatLit* ret = Expression::doubleToUnboxedFloatVal(v.toDouble());
+      if (ret) {
+        return ret;
+      }
+    }
+    
+    std::unordered_map<FloatVal, WeakRef>::iterator it = constants().floatMap.find(v);
+    if (it==constants().floatMap.end() || it->second()==NULL) {
+      FloatLit* fl = new FloatLit(Location().introduce(), v);
+      if (it==constants().floatMap.end()) {
+        constants().floatMap.insert(std::make_pair(v, fl));
+      } else {
+        it->second = fl;
+      }
+      return fl;
+    } else {
+      return it->second()->cast<FloatLit>();
+    }
+  }
+
   inline
   SetLit::SetLit(const Location& loc,
                  const std::vector<Expression*>& v)
-  : Expression(loc,E_SETLIT,Type()), _v(ASTExprVec<Expression>(v)), _isv(NULL) {
+  : Expression(loc,E_SETLIT,Type()), _v(ASTExprVec<Expression>(v)) {
+    _u.isv = NULL;
     rehash();
   }
   
   inline
   SetLit::SetLit(const Location& loc, ASTExprVec<Expression> v)
-  : Expression(loc,E_SETLIT,Type()), _v(v), _isv(NULL) {
+  : Expression(loc,E_SETLIT,Type()), _v(v) {
+    _u.isv = NULL;
     rehash();
   }
 
   inline
   SetLit::SetLit(const Location& loc, IntSetVal* isv)
-  : Expression(loc,E_SETLIT,Type()), _isv(isv) {
+  : Expression(loc,E_SETLIT,Type()) {
     _type = Type::parsetint();
+    _u.isv = isv;
+    rehash();
+  }
+
+  inline
+  SetLit::SetLit(const Location& loc, FloatSetVal* fsv)
+  : Expression(loc,E_SETLIT,Type()) {
+    _type = Type::parsetfloat();
+    _u.fsv = fsv;
     rehash();
   }
 
@@ -157,49 +311,71 @@ namespace MiniZinc {
   : Expression(loc,E_ANON,Type()) {
     rehash();
   }
-  
+
+
   inline
   ArrayLit::ArrayLit(const Location& loc,
-                     const std::vector<Expression*>& v,
+                     ArrayLit& v,
                      const std::vector<std::pair<int,int> >& dims)
   : Expression(loc,E_ARRAYLIT,Type()) {
     _flag_1 = false;
-    std::vector<int> d(dims.size()*2);
-    for (unsigned int i=dims.size(); i--;) {
-      d[i*2] = dims[i].first;
-      d[i*2+1] = dims[i].second;
+    _flag_2 = v._flag_2;
+    if (_flag_2) {
+      _u._al = v._u._al;
+      std::vector<int> d(dims.size()*2+v._dims.size()-v.dims()*2);
+      for (unsigned int i=static_cast<unsigned int>(dims.size()); i--;) {
+        d[i*2] = dims[i].first;
+        d[i*2+1] = dims[i].second;
+      }
+      int sliceOffset = static_cast<int>(dims.size())*2;
+      int origSliceOffset = v.dims()*2;
+      for (int i=0; i<_u._al->dims()*2; i++) {
+        d[sliceOffset+i] = v._dims[origSliceOffset+i];
+      }
+      _dims = ASTIntVec(d);
+    } else {
+      std::vector<int> d(dims.size()*2);
+      for (unsigned int i=static_cast<unsigned int>(dims.size()); i--;) {
+        d[i*2] = dims[i].first;
+        d[i*2+1] = dims[i].second;
+      }
+      if (v._u._v->flag() || d.size()!=2 || d[0]!=1) {
+        // only allocate dims vector if it is not a 1d array indexed from 1
+        _dims = ASTIntVec(d);
+      }
+      _u._v = v._u._v;
     }
-    _v = ASTExprVec<Expression>(v);
-    _dims = ASTIntVec(d);
     rehash();
   }
 
   inline
   ArrayLit::ArrayLit(const Location& loc,
-                     ASTExprVec<Expression> v,
-                     const std::vector<std::pair<int,int> >& dims)
+                     ArrayLit& v)
   : Expression(loc,E_ARRAYLIT,Type()) {
     _flag_1 = false;
-    std::vector<int> d(dims.size()*2);
-    for (unsigned int i=dims.size(); i--;) {
-      d[i*2] = dims[i].first;
-      d[i*2+1] = dims[i].second;
+    _flag_2 = v._flag_2;
+    if (_flag_2) {
+      _u._al = v._u._al;
+      std::vector<int> d(2+v._dims.size()-v.dims()*2);
+      d[0] = 1;
+      d[1] = v.size();
+      int sliceOffset = 2;
+      int origSliceOffset = v.dims()*2;
+      for (int i=0; i<_u._al->dims()*2; i++) {
+        d[sliceOffset+i] = v._dims[origSliceOffset+i];
+      }
+      _dims = ASTIntVec(d);
+    } else {
+      _u._v = v._u._v;
+      if (_u._v->flag()) {
+        std::vector<int> d(2);
+        d[0] = 1;
+        d[1] = v.length();
+        _dims = ASTIntVec(d);
+      } else {
+        // don't allocate dims vector since this is a 1d array indexed from 1
+      }
     }
-    _v = v;
-    _dims = ASTIntVec(d);
-    rehash();
-  }
-
-  inline
-  ArrayLit::ArrayLit(const Location& loc,
-                     ASTExprVec<Expression> v)
-  : Expression(loc,E_ARRAYLIT,Type()) {
-    _flag_1 = false;
-    std::vector<int> dims(2);
-    dims[0]=1;
-    dims[1]=v.size();
-    _v = v;
-    _dims = ASTIntVec(dims);
     rehash();
   }
 
@@ -208,11 +384,28 @@ namespace MiniZinc {
                      const std::vector<Expression*>& v)
   : Expression(loc,E_ARRAYLIT,Type()) {
     _flag_1 = false;
-    std::vector<int> dims(2);
-    dims[0]=1;
-    dims[1]=v.size();
-    _v = ASTExprVec<Expression>(v);
-    _dims = ASTIntVec(dims);
+    _flag_2 = false;
+    std::vector<int> d(2);
+    d[0] = 1;
+    d[1] = static_cast<int>(v.size());
+    compress(v, d);
+    rehash();
+  }
+
+  inline
+  ArrayLit::ArrayLit(const Location& loc,
+                     const std::vector<KeepAlive>& v)
+  : Expression(loc,E_ARRAYLIT,Type()) {
+    _flag_1 = false;
+    _flag_2 = false;
+    std::vector<int> d(2);
+    d[0] = 1;
+    d[1] = static_cast<int>(v.size());
+    std::vector<Expression*> vv(v.size());
+    for (unsigned int i=0; i<v.size(); i++) {
+      vv[i] = v[i]();
+    }
+    compress(vv, d);
     rehash();
   }
 
@@ -221,20 +414,20 @@ namespace MiniZinc {
                      const std::vector<std::vector<Expression*> >& v)
   : Expression(loc,E_ARRAYLIT,Type()) {
     _flag_1 = false;
+    _flag_2 = false;
     std::vector<int> dims(4);
     dims[0]=1;
-    dims[1]=v.size();
+    dims[1]=static_cast<int>(v.size());
     dims[2]=1;
-    dims[3]=v.size() > 0 ? v[0].size() : 0;
+    dims[3]=v.size() > 0 ? static_cast<int>(v[0].size()) : 0;
     std::vector<Expression*> vv;
     for (unsigned int i=0; i<v.size(); i++)
       for (unsigned int j=0; j<v[i].size(); j++)
         vv.push_back(v[i][j]);
-    _v = ASTExprVec<Expression>(vv);
-    _dims = ASTIntVec(dims);
+    compress(vv, dims);
     rehash();
   }
-
+  
   inline
   ArrayAccess::ArrayAccess(const Location& loc,
                            Expression* v,
@@ -261,16 +454,16 @@ namespace MiniZinc {
     std::vector<Expression*> es;
     std::vector<int> idx;
     for (unsigned int i=0; i<g._g.size(); i++) {
-      idx.push_back(es.size());
+      idx.push_back(static_cast<int>(es.size()));
       es.push_back(g._g[i]._in);
+      es.push_back(g._g[i]._where);
       for (unsigned int j=0; j<g._g[i]._v.size(); j++) {
         es.push_back(g._g[i]._v[j]);
       }
     }
-    idx.push_back(es.size());
+    idx.push_back(static_cast<int>(es.size()));
     _g = ASTExprVec<Expression>(es);
     _g_idx = ASTIntVec(idx);
-    _where = g._w;
     rehash();
   }
   inline
@@ -279,8 +472,8 @@ namespace MiniZinc {
                                Generators& g,
                                bool set)
   : Expression(loc,E_COMP,Type()) {
-    init(e,g);
     _flag_1 = set;
+    init(e,g);
   }
   inline void
   ITE::init(const std::vector<Expression*>& e_if_then, Expression* e_else) {
@@ -310,29 +503,66 @@ namespace MiniZinc {
     rehash();
   }
 
+  inline bool
+  Call::hasId(void) const {
+    return (reinterpret_cast<ptrdiff_t>(_u_id._decl) & static_cast<ptrdiff_t>(1)) == 0;
+  }
+  
+  inline ASTString
+  Call::id(void) const {
+    return hasId() ? _u_id._id : decl()->id();
+  }
+  
+  inline void
+  Call::id(const ASTString& i) {
+    _u_id._id = i.aststr();
+    assert(hasId());
+    assert(decl()==NULL);
+  }
 
+  inline FunctionI*
+  Call::decl(void) const {
+    return hasId() ? NULL : reinterpret_cast<FunctionI*>(reinterpret_cast<ptrdiff_t>(_u_id._decl) & ~static_cast<ptrdiff_t>(1));
+  }
+  
+  inline void
+  Call::decl(FunctionI* f) {
+    assert(f != NULL);
+    _u_id._decl = reinterpret_cast<FunctionI*>(reinterpret_cast<ptrdiff_t>(f) | static_cast<ptrdiff_t>(1));
+  }
+  
   inline
   Call::Call(const Location& loc,
-             const std::string& id,
-             const std::vector<Expression*>& args,
-             FunctionI* decl)
+             const std::string& id0,
+             const std::vector<Expression*>& args)
   : Expression(loc, E_CALL,Type()) {
-    _id = ASTString(id);
-    _args = ASTExprVec<Expression>(args);
-    _decl = decl;
+    _flag_1 = false;
+    id(ASTString(id0));
+    if (args.size()==1) {
+      _u._oneArg = args[0]->isUnboxedVal() ? args[0] : args[0]->tag();
+    } else {
+      _u._args = ASTExprVec<Expression>(args).vec();
+    }
     rehash();
+    assert(hasId());
+    assert(decl() == NULL);
   }
 
   inline
   Call::Call(const Location& loc,
-             const ASTString& id,
-             const std::vector<Expression*>& args,
-             FunctionI* decl)
+             const ASTString& id0,
+             const std::vector<Expression*>& args)
   : Expression(loc, E_CALL,Type()) {
-    _id = id;
-    _args = ASTExprVec<Expression>(args);
-    _decl = decl;
+    _flag_1 = false;
+    id(ASTString(id0));
+    if (args.size()==1) {
+      _u._oneArg = args[0]->isUnboxedVal() ? args[0] : args[0]->tag();
+    } else {
+      _u._args = ASTExprVec<Expression>(args).vec();
+    }
     rehash();
+    assert(hasId());
+    assert(decl() == NULL);
   }
 
   inline
@@ -400,11 +630,12 @@ namespace MiniZinc {
 
   inline Expression*
   VarDecl::e(void) const {
-    return reinterpret_cast<Expression*>(reinterpret_cast<ptrdiff_t>(_e) & ~ static_cast<ptrdiff_t>(1));
+    return (_e==nullptr || _e->isUnboxedVal()) ? _e : _e->untag();
   }
 
   inline void
   VarDecl::e(Expression* rhs) {
+    assert(rhs==NULL || !rhs->isa<Id>() || rhs->cast<Id>() != _id);
     _e = rhs;
   }
   
@@ -426,14 +657,16 @@ namespace MiniZinc {
   }
   inline bool
   VarDecl::evaluated(void) const {
-    return reinterpret_cast<ptrdiff_t>(_e) & 1;
+    return _e->isUnboxedVal() || _e->isTagged();
   }
   inline void
   VarDecl::evaluated(bool t) {
-    if (t)
-      _e = reinterpret_cast<Expression*>(reinterpret_cast<ptrdiff_t>(_e) | 1);
-    else
-      _e = reinterpret_cast<Expression*>(reinterpret_cast<ptrdiff_t>(_e) & ~static_cast<ptrdiff_t>(1));
+    if (!_e->isUnboxedVal()) {
+      if (t)
+        _e = _e->tag();
+      else
+        _e = _e->untag();
+    }
   }
   inline void
   VarDecl::flat(VarDecl* vd) {
@@ -447,6 +680,7 @@ namespace MiniZinc {
                      Expression* domain)
   : Expression(loc,E_TI,type), _ranges(ranges), _domain(domain) {
     _flag_1 = false;
+    _flag_2 = false;
     rehash();
   }
 
@@ -456,6 +690,7 @@ namespace MiniZinc {
                      Expression* domain)
   : Expression(loc,E_TI,type), _domain(domain) {
     _flag_1 = false;
+    _flag_2 = false;
     rehash();
   }
 
@@ -525,6 +760,36 @@ namespace MiniZinc {
     _builtins.i = NULL;
     _builtins.s = NULL;
     _builtins.str = NULL;
+    _from_stdlib = (loc.filename() == "builtins.mzn" ||
+              loc.filename().endsWith("/builtins.mzn") ||
+              loc.filename() == "stdlib.mzn" ||
+              loc.filename().endsWith("/stdlib.mzn") ||
+              loc.filename() == "flatzinc_builtins.mzn" ||
+              loc.filename().endsWith("/flatzinc_builtins.mzn"));
   }
 
-}
\ No newline at end of file
+  inline
+  FunctionI::FunctionI(const Location& loc,
+                       const ASTString& id, TypeInst* ti,
+                       const ASTExprVec<VarDecl>& params,
+                       Expression* e)
+  : Item(loc, II_FUN),
+  _id(id),
+  _ti(ti),
+  _params(params),
+  _e(e) {
+    _builtins.e = NULL;
+    _builtins.b = NULL;
+    _builtins.f = NULL;
+    _builtins.i = NULL;
+    _builtins.s = NULL;
+    _builtins.str = NULL;
+    _from_stdlib = (loc.filename() == "builtins.mzn" ||
+                    loc.filename().endsWith("/builtins.mzn") ||
+                    loc.filename() == "stdlib.mzn" ||
+                    loc.filename().endsWith("/stdlib.mzn") ||
+                    loc.filename() == "flatzinc_builtins.mzn" ||
+                    loc.filename().endsWith("/flatzinc_builtins.mzn"));
+  }
+
+}
