@@ -14,6 +14,7 @@
 #include <minizinc/hash.hh>
 #include <minizinc/model.hh>
 #include <minizinc/config.hh>
+#include <minizinc/timer.hh>
 
 #include <vector>
 #include <cstring>
@@ -89,10 +90,14 @@ namespace MiniZinc {
     std::map<int,GCStat> gc_stats;
 #endif
   protected:
+    
+    static const size_t _min_gc_threshold;
+    
     HeapPage* _page;
     Model* _rootset;
     KeepAlive* _roots;
     WeakRef* _weakRefs;
+    ASTNodeWeakMap* _nodeWeakMaps;
     static const int _max_fl = 5;
     FreeListNode* _fl[_max_fl+1];
     static const size_t _fl_size[_max_fl+1];
@@ -133,9 +138,10 @@ namespace MiniZinc {
       , _rootset(NULL)
       , _roots(NULL)
       , _weakRefs(NULL)
+      , _nodeWeakMaps(NULL)
       , _alloced_mem(0)
       , _free_mem(0)
-      , _gc_threshold(10)
+      , _gc_threshold(_min_gc_threshold)
       , _max_alloced_mem(0) {
       for (int i=_max_fl+1; i--;)
         _fl[i] = NULL;
@@ -149,8 +155,11 @@ namespace MiniZinc {
         s = std::max(s,pageSize);
       HeapPage* newPage =
         static_cast<HeapPage*>(::malloc(sizeof(HeapPage)+s-1));
+      if (newPage==NULL) {
+        throw InternalError("out of memory");
+      }
 #ifndef NDEBUG
-        memset(newPage,255,sizeof(HeapPage)+s-1);
+      memset(newPage,255,sizeof(HeapPage)+s-1);
 #endif
       _alloced_mem += s;
       _max_alloced_mem = std::max(_max_alloced_mem, _alloced_mem);
@@ -192,6 +201,11 @@ namespace MiniZinc {
       char* ret = p->data+p->used;
       p->used += size;
       _free_mem -= size;
+      if (p->size-p->used < _fl_size[0]) {
+        _free_mem -= (p->size-p->used);
+        _alloced_mem -= (p->size-p->used);
+        p->size=p->used;
+      }
       assert(_alloced_mem >= _free_mem);
       return ret;
     }
@@ -216,23 +230,37 @@ namespace MiniZinc {
       return alloc(size);
     }
 
+    void trigger(void) {
+#ifdef MINIZINC_GC_STATS
+      std::cerr << "GC\n\talloced " << (_alloced_mem/1024) << "\n\tfree " << (_free_mem/1024) << "\n\tdiff "
+      << ((_alloced_mem-_free_mem)/1024)
+      << "\n\tthreshold " << (_gc_threshold/1024)
+      << "\n";
+#endif
+      size_t old_free = _free_mem;
+      mark();
+      sweep();
+      // GC strategy:
+      // increase threshold if either
+      //   a) we haven't been able to put much on the free list (comapred to before GC), or
+      //   b) the free list memory (after GC) is less than 50% of the allocated memory
+      // otherwise (i.e., we have been able to increase the free list, and it is now more
+      // than 50% of overall allocated memory), keep threshold at allocated memory
+      if ( (old_free!= 0 && static_cast<double>(old_free)/static_cast<double>(_free_mem) > 0.9) || static_cast<double>(_free_mem)/_alloced_mem < 0.5) {
+        _gc_threshold = std::max(_min_gc_threshold,static_cast<size_t>(_alloced_mem * 1.5));
+      } else {
+        _gc_threshold = std::max(_min_gc_threshold,_alloced_mem);
+      }
+#ifdef MINIZINC_GC_STATS
+      std::cerr << "done\n\talloced " << (_alloced_mem/1024) << "\n\tfree " << (_free_mem/1024) << "\n\tdiff "
+      << ((_alloced_mem-_free_mem)/1024)
+      << "\n\tthreshold " << (_gc_threshold/1024)
+      << "\n";
+#endif
+    }
     void rungc(void) {
       if (_alloced_mem > _gc_threshold) {
-#ifdef MINIZINC_GC_STATS
-        std::cerr << "GC\n\talloced " << (_alloced_mem/1024) << "\n\tfree " << (_free_mem/1024) << "\n\tdiff "
-                  << ((_alloced_mem-_free_mem)/1024)
-                  << "\n\tthreshold " << (_gc_threshold/1024)
-                  << "\n";
-#endif
-        mark();
-        sweep();
-        _gc_threshold = static_cast<size_t>(_alloced_mem * 1.5);
-#ifdef MINIZINC_GC_STATS
-        std::cerr << "done\n\talloced " << (_alloced_mem/1024) << "\n\tfree " << (_free_mem/1024) << "\n\tdiff "
-                  << ((_alloced_mem-_free_mem)/1024)
-                  << "\n\tthreshold " << (_gc_threshold/1024)
-                  << "\n";
-#endif
+        trigger();
       }
     }
     void mark(void);
@@ -281,6 +309,9 @@ namespace MiniZinc {
       case ASTNode::NID_VEC:
         ns = static_cast<ASTVec*>(n)->memsize();
         break;
+      case Call::eid:
+        ns = n->_flag_1 ? _nodesize[BinOp::eid] : _nodesize[Call::eid];
+        break;
       default:
         assert(n->_id <= Item::II_END);
         ns = _nodesize[n->_id];
@@ -293,6 +324,8 @@ namespace MiniZinc {
 
   };
 
+  const size_t GC::Heap::_min_gc_threshold = 10ll*1024ll;
+  
 #ifdef MINIZINC_GC_STATS
   const char*
   GC::Heap::_nodeid[] = {
@@ -327,13 +360,34 @@ namespace MiniZinc {
   };
 #endif
   
+  void
+  GC::setTimeout(unsigned long long int t) {
+    if (gc()==NULL) {
+      gc() = new GC();
+    }
+    gc()->_timeout = t;
+    gc()->_timeout_timer.reset();
+  }
+  
   void
   GC::lock(void) {
     if (gc()==NULL) {
       gc() = new GC();
     }
-    if (gc()->_lock_count==0)
+    // If a timeout has been specified, first check counter
+    // before checking timer (counter is much cheaper, introduces
+    // less overhead)
+    if (gc()->_timeout > 0 && gc()->_timeout_counter++ > 500) {
+      gc()->_timeout_counter = 0;
+      if (gc()->_timeout_timer.ms() > gc()->_timeout) {
+        gc()->_timeout = 0;
+        gc()->_timeout_counter = 0;
+        throw Timeout();
+      }
+    }
+    if (gc()->_lock_count==0) {
       gc()->_heap->rungc();
+    }
     gc()->_lock_count++;
   }
   void
@@ -342,6 +396,11 @@ namespace MiniZinc {
     gc()->_lock_count--;
   }
 
+  void GC::trigger(void) {
+    if (!locked())
+      gc()->_heap->trigger();
+  }
+  
   const size_t GC::Heap::pageSize;
 
   const size_t
@@ -354,7 +413,7 @@ namespace MiniZinc {
     sizeof(Item)+6*sizeof(void*),
   };
 
-  GC::GC(void) : _heap(new Heap()), _lock_count(0) {}
+  GC::GC(void) : _heap(new Heap()), _lock_count(0), _timeout(0), _timeout_counter(0) {}
 
   void
   GC::add(Model* m) {
@@ -479,23 +538,40 @@ namespace MiniZinc {
       m = m->_roots_next;
     } while (m != _rootset);
     
-    for (unsigned int i=trail.size(); i--;) {
+    for (unsigned int i=static_cast<unsigned int>(trail.size()); i--;) {
       Expression::mark(trail[i].v);
     }
     
     bool fixPrev = false;
+    WeakRef* prevWr = NULL;
     for (WeakRef* wr = _weakRefs; wr != NULL; wr = wr->next()) {
       if (fixPrev) {
         fixPrev = false;
-        WeakRef* p = wr->_p;
-        removeWeakRef(p);
-        p->_n = p;
-        p->_p = p;
+        removeWeakRef(prevWr);
+        prevWr->_n = NULL;
+        prevWr->_p = NULL;
       }
       if ((*wr)() && (*wr)()->_gc_mark==0) {
         wr->_e = NULL;
         wr->_valid = false;
         fixPrev = true;
+        prevWr = wr;
+      }
+    }
+    if (fixPrev) {
+      removeWeakRef(prevWr);
+      prevWr->_n = NULL;
+      prevWr->_p = NULL;
+    }
+    
+    for (ASTNodeWeakMap* wr = _nodeWeakMaps; wr != NULL; wr = wr->next()) {
+      std::vector<ASTNode*> toRemove;
+      for (auto n : wr->_m) {
+        if (n.first->_gc_mark==0 || n.second->_gc_mark==0)
+          toRemove.push_back(n.first);
+      }
+      for (auto n : toRemove) {
+        wr->_m.erase(n);
       }
     }
 
@@ -514,7 +590,10 @@ namespace MiniZinc {
     HeapPage* prev = NULL;
     while (p) {
       size_t off = 0;
-      bool wholepage = false;
+      bool wholepage = true;
+      struct NodeInfo { ASTNode* n; size_t ns; NodeInfo(ASTNode* n0, size_t ns0) : n(n0), ns(ns0) {} };
+      std::vector<NodeInfo> freeNodes;
+      freeNodes.reserve(100);
       while (off < p->used) {
         ASTNode* n = reinterpret_cast<ASTNode*>(p->data+off);
         size_t ns = nodesize(n);
@@ -542,23 +621,16 @@ namespace MiniZinc {
               }
           }
           if (ns >= _fl_size[0] && ns <= _fl_size[_max_fl]) {
-            FreeListNode* fln = static_cast<FreeListNode*>(n);
-            new (fln) FreeListNode(ns, _fl[_fl_slot(ns)]);
-            _fl[_fl_slot(ns)] = fln;
-            _free_mem += ns;
-#if defined(MINIZINC_GC_STATS)
-            gc_stats[fln->_id].second++;
-#endif
-            assert(_alloced_mem >= _free_mem);
+            freeNodes.push_back(NodeInfo(n,ns));
           } else {
             assert(off==0);
             assert(p->used==p->size);
-            wholepage = true;
           }
         } else {
 #if defined(MINIZINC_GC_STATS)
           stats.second++;
 #endif
+          wholepage = false;
           if (n->_id != ASTNode::NID_FL)
             n->_gc_mark=0;
         }
@@ -577,9 +649,21 @@ namespace MiniZinc {
         HeapPage* pf = p;
         p = p->next;
         _alloced_mem -= pf->size;
+        if (pf->size-pf->used >= _fl_size[0])
+          _free_mem -= (pf->size-pf->used);
         assert(_alloced_mem >= _free_mem);
         ::free(pf);
       } else {
+        for (auto ni : freeNodes) {
+          FreeListNode* fln = static_cast<FreeListNode*>(ni.n);
+          new (fln) FreeListNode(ni.ns, _fl[_fl_slot(ni.ns)]);
+          _fl[_fl_slot(ni.ns)] = fln;
+          _free_mem += ni.ns;
+#if defined(MINIZINC_GC_STATS)
+          gc_stats[fln->_id].second++;
+#endif
+        }
+        assert(_alloced_mem >= _free_mem);
         prev = p;
         p = p->next;
       }
@@ -677,24 +761,26 @@ namespace MiniZinc {
 
   KeepAlive::KeepAlive(Expression* e)
     : _e(e), _p(NULL), _n(NULL) {
-    if (_e)
+    if (_e && !_e->isUnboxedVal())
       GC::gc()->addKeepAlive(this);
   }
   KeepAlive::~KeepAlive(void) {
-    if (_e)
+    if (_e && !_e->isUnboxedVal())
       GC::gc()->removeKeepAlive(this);
   }
   KeepAlive::KeepAlive(const KeepAlive& e) : _e(e._e), _p(NULL), _n(NULL) {
-    if (_e)
+    if (_e && !_e->isUnboxedVal())
       GC::gc()->addKeepAlive(this);
   }
   KeepAlive&
   KeepAlive::operator =(const KeepAlive& e) {
-    if (_e) {
-      if (e._e==NULL)
+    if (_e && !_e->isUnboxedVal()) {
+      if (e._e==NULL || e._e->isUnboxedVal()) {
         GC::gc()->removeKeepAlive(this);
+        _p = _n = NULL;
+      }
     } else {
-      if (e._e!=NULL)
+      if (e._e!=NULL && !e._e->isUnboxedVal())
         GC::gc()->addKeepAlive(this);
     }
     _e = e._e;
@@ -722,34 +808,81 @@ namespace MiniZinc {
       e->_n->_p = e->_p;
     }
   }
+  void
+  GC::addNodeWeakMap(ASTNodeWeakMap* m) {
+    assert(m->_p==NULL);
+    assert(m->_n==NULL);
+    m->_n = GC::gc()->_heap->_nodeWeakMaps;
+    if (GC::gc()->_heap->_nodeWeakMaps)
+      GC::gc()->_heap->_nodeWeakMaps->_p = m;
+    GC::gc()->_heap->_nodeWeakMaps = m;
+  }
+  void
+  GC::removeNodeWeakMap(ASTNodeWeakMap* m) {
+    if (m->_p) {
+      m->_p->_n = m->_n;
+    } else {
+      assert(GC::gc()->_heap->_nodeWeakMaps==m);
+      GC::gc()->_heap->_nodeWeakMaps = m->_n;
+    }
+    if (m->_n) {
+      m->_n->_p = m->_p;
+    }
+  }
 
   WeakRef::WeakRef(Expression* e)
   : _e(e), _p(NULL), _n(NULL), _valid(true) {
-    if (_e)
+    if (_e && !_e->isUnboxedVal())
       GC::gc()->addWeakRef(this);
   }
   WeakRef::~WeakRef(void) {
-    if (_e || !_valid)
+    if (_e && !_e->isUnboxedVal())
       GC::gc()->removeWeakRef(this);
   }
   WeakRef::WeakRef(const WeakRef& e) : _e(e()), _p(NULL), _n(NULL), _valid(true) {
-    if (_e)
+    if (_e && !_e->isUnboxedVal())
       GC::gc()->addWeakRef(this);
   }
   WeakRef&
   WeakRef::operator =(const WeakRef& e) {
-    if (_e || !_valid) {
-      if (e()==NULL) {
+    // Test if this WeakRef is currently active in the GC
+    bool isActive = (_e && !_e->isUnboxedVal());
+    if (isActive) {
+      // Yes, active WeakRef.
+      // If after assigning WeakRef should be inactive, remove it.
+      if (e()==NULL || e()->isUnboxedVal()) {
         GC::gc()->removeWeakRef(this);
         _n = _p = NULL;
       }
-    } else {
-      if (e()!=NULL)
-        GC::gc()->addWeakRef(this);
     }
     _e = e();
     _valid = true;
+    
+    // If this WeakRef was not active but now should be, add it
+    if (!isActive && _e!=NULL && !_e->isUnboxedVal())
+      GC::gc()->addWeakRef(this);
     return *this;
   }
 
+  ASTNodeWeakMap::ASTNodeWeakMap(void)
+  : _p(NULL), _n(NULL) {
+    GC::gc()->addNodeWeakMap(this);
+  }
+  
+  ASTNodeWeakMap::~ASTNodeWeakMap(void) {
+    GC::gc()->removeNodeWeakMap(this);
+  }
+
+  void
+  ASTNodeWeakMap::insert(ASTNode* n0, ASTNode* n1) {
+    _m.insert(std::make_pair(n0,n1));
+  }
+  
+  ASTNode*
+  ASTNodeWeakMap::find(ASTNode* n) {
+    NodeMap::iterator it = _m.find(n);
+    if (it==_m.end()) return NULL;
+    return it->second;
+  }
+  
 }
