@@ -38,26 +38,26 @@ namespace MiniZinc {
   namespace Optimizers {
     
     OptimizeRegistry::ConstraintStatus o_linear(EnvI& env, Item* ii, Call* c, Expression*& rewrite) {
-      ArrayLit* al_c = eval_array_lit(env,c->args()[0]);
-      std::vector<IntVal> coeffs(al_c->v().size());
-      for (unsigned int i=0; i<al_c->v().size(); i++) {
-        coeffs[i] = eval_int(env,al_c->v()[i]);
+      ArrayLit* al_c = eval_array_lit(env,c->arg(0));
+      std::vector<IntVal> coeffs(al_c->size());
+      for (unsigned int i=0; i<al_c->size(); i++) {
+        coeffs[i] = eval_int(env,(*al_c)[i]);
       }
-      ArrayLit* al_x = eval_array_lit(env,c->args()[1]);
-      std::vector<KeepAlive> x(al_x->v().size());
-      for (unsigned int i=0; i<al_x->v().size(); i++) {
-        x[i] = al_x->v()[i];
+      ArrayLit* al_x = eval_array_lit(env,c->arg(1));
+      std::vector<KeepAlive> x(al_x->size());
+      for (unsigned int i=0; i<al_x->size(); i++) {
+        x[i] = (*al_x)[i];
       }
       IntVal d = 0;
       simplify_lin<IntLit>(coeffs, x, d);
       if (coeffs.size()==0) {
         bool failed;
         if (c->id()==constants().ids.int_.lin_le) {
-          failed = (d > eval_int(env,c->args()[2]));
+          failed = (d > eval_int(env,c->arg(2)));
         } else if (c->id()==constants().ids.int_.lin_eq) {
-          failed = (d != eval_int(env,c->args()[2]));
+          failed = (d != eval_int(env,c->arg(2)));
         } else {
-          failed = (d == eval_int(env,c->args()[2]));
+          failed = (d == eval_int(env,c->arg(2)));
         }
         if (failed) {
           return OptimizeRegistry::CS_FAILED;
@@ -68,7 +68,7 @@ namespace MiniZinc {
         VarDecl* vd = x[0]()->cast<Id>()->decl();
         IntSetVal* domain = vd->ti()->domain() ? eval_intset(env,vd->ti()->domain()) : NULL;
         if (c->id()==constants().ids.int_.lin_eq) {
-          IntVal rd = eval_int(env,c->args()[2])-d;
+          IntVal rd = eval_int(env,c->arg(2))-d;
           if (rd % coeffs[0] == 0) {
             IntVal nd = rd / coeffs[0];
             if (domain && !domain->contains(nd))
@@ -84,7 +84,7 @@ namespace MiniZinc {
           }
         } else if (c->id()==constants().ids.int_.lin_le) {
           IntVal ac = std::abs(coeffs[0]);
-          IntVal rd = eval_int(env,c->args()[2])-d;
+          IntVal rd = eval_int(env,c->arg(2))-d;
           IntVal ad = std::abs(rd);
           IntVal nd;
           if (ad % ac == 0) {
@@ -127,14 +127,14 @@ namespace MiniZinc {
           }
         }
       } else if (c->id()==constants().ids.int_.lin_eq && coeffs.size()==2  &&
-                 ((coeffs[0]==1 && coeffs[1]==-1) || (coeffs[1]==1 && coeffs[0]==-1)) && eval_int(env,c->args()[2])-d==0) {
+                 ((coeffs[0]==1 && coeffs[1]==-1) || (coeffs[1]==1 && coeffs[0]==-1)) && eval_int(env,c->arg(2))-d==0) {
         std::vector<Expression*> args(2);
         args[0] = x[0](); args[1] = x[1]();
         Call* c = new Call(Location(), constants().ids.int_.eq, args);
         rewrite = c;
         return OptimizeRegistry::CS_REWRITE;
       }
-      if (coeffs.size() < al_c->v().size()) {
+      if (coeffs.size() < al_c->size()) {
         std::vector<Expression*> coeffs_e(coeffs.size());
         std::vector<Expression*> x_e(coeffs.size());
         for (unsigned int i=0; i<coeffs.size(); i++) {
@@ -149,7 +149,7 @@ namespace MiniZinc {
         std::vector<Expression*> args(3);
         args[0] = al_c_new;
         args[1] = al_x_new;
-        args[2] = IntLit::a(eval_int(env,c->args()[2])-d);
+        args[2] = IntLit::a(eval_int(env,c->arg(2))-d);
         Call* nc = new Call(Location(), c->id(), args);
         nc->type(Type::varbool());
         for (ExpressionSetIter it = c->ann().begin(); it != c->ann().end(); ++it) {
@@ -164,22 +164,22 @@ namespace MiniZinc {
 
     OptimizeRegistry::ConstraintStatus o_lin_exp(EnvI& env, Item* i, Call* c, Expression*& rewrite) {
       if (c->type().isint()) {
-        ArrayLit* al_c = eval_array_lit(env,c->args()[0]);
-        std::vector<IntVal> coeffs(al_c->v().size());
-        for (unsigned int i=0; i<al_c->v().size(); i++) {
-          coeffs[i] = eval_int(env,al_c->v()[i]);
+        ArrayLit* al_c = eval_array_lit(env,c->arg(0));
+        std::vector<IntVal> coeffs(al_c->size());
+        for (unsigned int i=0; i<al_c->size(); i++) {
+          coeffs[i] = eval_int(env,(*al_c)[i]);
         }
-        ArrayLit* al_x = eval_array_lit(env,c->args()[1]);
-        std::vector<KeepAlive> x(al_x->v().size());
-        for (unsigned int i=0; i<al_x->v().size(); i++) {
-          x[i] = al_x->v()[i];
+        ArrayLit* al_x = eval_array_lit(env,c->arg(1));
+        std::vector<KeepAlive> x(al_x->size());
+        for (unsigned int i=0; i<al_x->size(); i++) {
+          x[i] = (*al_x)[i];
         }
-        IntVal d = eval_int(env,c->args()[2]);
+        IntVal d = eval_int(env,c->arg(2));
         simplify_lin<IntLit>(coeffs, x, d);
         if (coeffs.size()==0) {
           rewrite = IntLit::a(d);
           return OptimizeRegistry::CS_REWRITE;
-        } else if (coeffs.size() < al_c->v().size()) {
+        } else if (coeffs.size() < al_c->size()) {
           if (coeffs.size()==1 && coeffs[0]==1 && d==0) {
             rewrite = x[0]();
             return OptimizeRegistry::CS_REWRITE;
@@ -214,16 +214,16 @@ namespace MiniZinc {
 
     
     OptimizeRegistry::ConstraintStatus o_element(EnvI& env, Item* i, Call* c, Expression*& rewrite) {
-      if (c->args()[0]->isa<IntLit>()) {
-        IntVal idx = eval_int(env,c->args()[0]);
-        ArrayLit* al = eval_array_lit(env,c->args()[1]);
-        if (idx < 1 || idx > al->v().size()) {
+      if (c->arg(0)->isa<IntLit>()) {
+        IntVal idx = eval_int(env,c->arg(0));
+        ArrayLit* al = eval_array_lit(env,c->arg(1));
+        if (idx < 1 || idx > al->size()) {
           return OptimizeRegistry::CS_FAILED;
         }
-        Expression* result = al->v()[idx.toInt()-1];
+        Expression* result = (*al)[static_cast<int>(idx.toInt())-1];
         std::vector<Expression*> args(2);
         args[0] = result;
-        args[1] = c->args()[2];
+        args[1] = c->arg(2);
         Call* eq = new Call(Location(),constants().ids.int_.eq,args);
         rewrite = eq;
         return OptimizeRegistry::CS_REWRITE;
@@ -234,16 +234,16 @@ namespace MiniZinc {
     OptimizeRegistry::ConstraintStatus o_clause(EnvI& env, Item* i, Call* c, Expression*& rewrite) {
       std::vector<VarDecl*> pos;
       std::vector<VarDecl*> neg;
-      ArrayLit* al_pos = eval_array_lit(env, c->args()[0]);
-      for (unsigned int i=0; i<al_pos->v().size(); i++) {
-        if (Id* ident = al_pos->v()[i]->dyn_cast<Id>()) {
+      ArrayLit* al_pos = eval_array_lit(env, c->arg(0));
+      for (unsigned int i=0; i<al_pos->size(); i++) {
+        if (Id* ident = (*al_pos)[i]->dyn_cast<Id>()) {
           if (ident->decl()->ti()->domain()==NULL)
             pos.push_back(ident->decl());
         }
       }
-      ArrayLit* al_neg = eval_array_lit(env, c->args()[1]);
-      for (unsigned int i=0; i<al_neg->v().size(); i++) {
-        if (Id* ident = al_neg->v()[i]->dyn_cast<Id>()) {
+      ArrayLit* al_neg = eval_array_lit(env, c->arg(1));
+      for (unsigned int i=0; i<al_neg->size(); i++) {
+        if (Id* ident = (*al_neg)[i]->dyn_cast<Id>()) {
           if (ident->decl()->ti()->domain()==NULL)
             neg.push_back(ident->decl());
         }
@@ -274,29 +274,186 @@ namespace MiniZinc {
         return OptimizeRegistry::CS_OK;
       }
     }
-    
+
+    OptimizeRegistry::ConstraintStatus o_div(EnvI& env, Item* i, Call* c, Expression*& rewrite) {
+      if (c->arg(1)->type().ispar()) {
+        IntVal c1v = eval_int(env, c->arg(1));
+        if (c->arg(0)->type().ispar() && c->n_args()==3 && c->arg(2)->type().ispar()) {
+          IntVal c0v = eval_int(env, c->arg(0));
+          IntVal c2v = eval_int(env, c->arg(2));
+          return (c0v / c1v == c2v) ? OptimizeRegistry::CS_ENTAILED : OptimizeRegistry::CS_FAILED;
+        }
+      }
+      return OptimizeRegistry::CS_OK;
+    }
+
+    OptimizeRegistry::ConstraintStatus o_times(EnvI& env, Item* i, Call* c, Expression*& rewrite) {
+      Expression* result = nullptr;
+      Expression* arg0 = c->arg(0);
+      Expression* arg1 = c->arg(1);
+      if (arg0->type().ispar() && arg1->type().ispar()) {
+        IntVal c0v = eval_int(env, arg0);
+        IntVal c1v = eval_int(env, arg1);
+        result = IntLit::a(c0v*c1v);
+      } else if (arg0->type().ispar()) {
+        IntVal c0v = eval_int(env, arg0);
+        if (c0v==0) {
+          result = IntLit::a(0);
+        } else if (c0v==1) {
+          result = arg1;
+        }
+      } else if (arg1->type().ispar()) {
+        IntVal c1v = eval_int(env, arg1);
+        if (c1v==0) {
+          result = IntLit::a(0);
+        }
+        if (c1v==1) {
+          result = arg0;
+        }
+      }
+
+      if (result) {
+        if (c->n_args()==2) {
+          // this is the functional version of times
+          rewrite = result;
+          return OptimizeRegistry::CS_REWRITE;
+        } else {
+          // this is the relational version of times
+          assert(c->n_args()==3);
+          rewrite = new Call(Location().introduce(), constants().ids.int_.eq, {c->arg(2),result});
+          return OptimizeRegistry::CS_REWRITE;
+        }
+      }
+      return OptimizeRegistry::CS_OK;
+    }
+
+    OptimizeRegistry::ConstraintStatus o_set_in(EnvI& env, Item* i, Call* c, Expression*& rewrite) {
+      if (c->arg(1)->type().ispar()) {
+        if (c->arg(0)->type().ispar()) {
+          IntSetVal* isv = eval_intset(env, c->arg(1));
+          return isv->contains(eval_int(env,c->arg(0))) ? OptimizeRegistry::CS_ENTAILED : OptimizeRegistry::CS_FAILED;
+        } else if (Id* ident = c->arg(0)->dyn_cast<Id>()) {
+          VarDecl* vd = ident->decl();
+          IntSetVal* isv = eval_intset(env, c->arg(1));
+          if (vd->ti()->domain()) {
+            IntSetVal* dom = eval_intset(env, vd->ti()->domain());
+            {
+              IntSetRanges isv_r(isv);
+              IntSetRanges dom_r(dom);
+              if (Ranges::subset(dom_r, isv_r))
+                return OptimizeRegistry::CS_ENTAILED;
+            }
+            {
+              IntSetRanges isv_r(isv);
+              IntSetRanges dom_r(dom);
+              if (Ranges::disjoint(dom_r, isv_r))
+                return OptimizeRegistry::CS_FAILED;
+            }
+          } else if (isv->min()==isv->max()) {
+            std::vector<Expression*> args(2);
+            args[0] = vd->id();
+            args[1] = IntLit::a(isv->min());
+            Call* eq = new Call(Location(),constants().ids.int_.eq,args);
+            rewrite = eq;
+            return OptimizeRegistry::CS_REWRITE;
+            
+          }
+        }
+      }
+      return OptimizeRegistry::CS_OK;
+    }
+
+    OptimizeRegistry::ConstraintStatus o_int_ne(EnvI& env, Item* i, Call* c, Expression*& rewrite) {
+      Expression* e0 = c->arg(0);
+      Expression* e1 = c->arg(1);
+      if (e0->type().ispar() && e1->type().ispar()) {
+        return eval_int(env, e0) != eval_int(env, e1) ? OptimizeRegistry::CS_ENTAILED : OptimizeRegistry::CS_FAILED;
+      }
+      if (e1->isa<Id>()) {
+        std::swap(e0, e1);
+      }
+      if (Id* ident = e0->dyn_cast<Id>()) {
+        if (e1->type().ispar()) {
+          if (ident->decl()->ti()->domain()) {
+            IntVal e1v = eval_int(env, e1);
+            IntSetVal* isv = eval_intset(env, ident->decl()->ti()->domain());
+            if (!isv->contains(e1v))
+              return OptimizeRegistry::CS_ENTAILED;
+            if (e1v==isv->min() && e1v==isv->max())
+              return OptimizeRegistry::CS_FAILED;
+          }
+        }
+      }
+      
+      return OptimizeRegistry::CS_OK;
+    }
+
+    OptimizeRegistry::ConstraintStatus o_int_le(EnvI& env, Item* i, Call* c, Expression*& rewrite) {
+      Expression* e0 = c->arg(0);
+      Expression* e1 = c->arg(1);
+      if (e0->type().ispar() && e1->type().ispar()) {
+        return eval_int(env, e0) <= eval_int(env, e1) ? OptimizeRegistry::CS_ENTAILED : OptimizeRegistry::CS_FAILED;
+      }
+      bool swapped = false;
+      if (e1->isa<Id>()) {
+        std::swap(e0, e1);
+        swapped = true;
+      }
+      if (Id* ident = e0->dyn_cast<Id>()) {
+        if (e1->type().ispar()) {
+          if (ident->decl()->ti()->domain()) {
+            IntVal e1v = eval_int(env, e1);
+            IntSetVal* isv = eval_intset(env, ident->decl()->ti()->domain());
+            if (!swapped) {
+              if (isv->max() <= e1v)
+                return OptimizeRegistry::CS_ENTAILED;
+              if (isv->min() > e1v)
+                return OptimizeRegistry::CS_FAILED;
+            } else {
+              if (e1v <= isv->min())
+                return OptimizeRegistry::CS_ENTAILED;
+              if (e1v > isv->max())
+                return OptimizeRegistry::CS_FAILED;
+            }
+          }
+        }
+      }
+      
+      return OptimizeRegistry::CS_OK;
+    }
+
     class Register {
+    private:
+      Model* _keepAliveModel;
     public:
       Register(void) {
         GCLock lock;
-        Model* m = new Model;
+        _keepAliveModel = new Model;
         ASTString id_element("array_int_element");
         ASTString id_var_element("array_var_int_element");
         std::vector<Expression*> e;
         e.push_back(new StringLit(Location(),id_element));
         e.push_back(new StringLit(Location(),id_var_element));
-        m->addItem(new ConstraintI(Location(),new ArrayLit(Location(),e)));
+        _keepAliveModel->addItem(new ConstraintI(Location(),new ArrayLit(Location(),e)));
         OptimizeRegistry::registry().reg(constants().ids.int_.lin_eq, o_linear);
         OptimizeRegistry::registry().reg(constants().ids.int_.lin_le, o_linear);
         OptimizeRegistry::registry().reg(constants().ids.int_.lin_ne, o_linear);
+        OptimizeRegistry::registry().reg(constants().ids.int_.div, o_div);
+        OptimizeRegistry::registry().reg(constants().ids.int_.times, o_times);
         OptimizeRegistry::registry().reg(id_element, o_element);
         OptimizeRegistry::registry().reg(constants().ids.lin_exp, o_lin_exp);
         OptimizeRegistry::registry().reg(id_var_element, o_element);
         OptimizeRegistry::registry().reg(constants().ids.clause, o_clause);
         OptimizeRegistry::registry().reg(constants().ids.bool_clause, o_clause);
+        OptimizeRegistry::registry().reg(constants().ids.set_in, o_set_in);
+        OptimizeRegistry::registry().reg(constants().ids.int_.ne, o_int_ne);
+        OptimizeRegistry::registry().reg(constants().ids.int_.le, o_int_le);
+      }
+      ~Register(void) {
+        delete _keepAliveModel;
       }
     } _r;
     
   }
   
-}
\ No newline at end of file
+}
