@@ -79,6 +79,8 @@ namespace MiniZinc {
         return 300;
       case BOT_INTERSECT:
         return 300;
+      case BOT_POW:
+        return 200;
       case BOT_PLUSPLUS:
         return 200;
       default:
@@ -134,36 +136,57 @@ namespace MiniZinc {
   
   std::string Printer::escapeStringLit(const ASTString& s) {
     const char* sc = s.c_str();
-    std::string ret;
+    std::ostringstream ret;
     for (unsigned int i=0; i<s.size(); i++) {
       switch (sc[i]) {
         case '\n':
-          ret += "\\n";
+          ret << "\\n";
           break;
         case '\t':
-          ret += "\\t";
+          ret << "\\t";
           break;
         case '"':
-          ret += "\\\"";
-          break;
-        case '\'':
-          ret += "\\\'";
+          ret << "\\\"";
           break;
         case '\\':
-          ret += "\\\\";
+          ret << "\\\\";
           break;
         default:
-          ret += sc[i];
+          ret << sc[i];
       }
     }
-    return ret;
+    return ret.str();
+  }
+  
+  void ppFloatVal(std::ostream& os, const FloatVal& fv, bool hexFloat) {
+    std::ostringstream oss;
+    if (fv.isFinite()) {
+      if (hexFloat) {
+        throw InternalError( "disabled due to hexfloat being not supported by g++ 4.9" );
+//          std::hexfloat(oss);
+        oss << fv.toDouble();
+        os << oss.str();
+      } else {
+        oss << std::setprecision(std::numeric_limits<double>::digits10+1);
+        oss << fv;
+        if (oss.str().find("e") == std::string::npos && oss.str().find(".") == std::string::npos)
+          oss << ".0";
+        os << oss.str();
+      }
+    } else {
+      if (fv.isPlusInfinity())
+        os << "infinity";
+      else
+        os << "-infinity";
+    }
   }
   
   class PlainPrinter {
   public:
+    EnvI* env;
     std::ostream& os;
     bool _flatZinc;
-    PlainPrinter(std::ostream& os0, bool flatZinc) : os(os0), _flatZinc(flatZinc) {}
+    PlainPrinter(std::ostream& os0, bool flatZinc, EnvI* env0) : env(env0), os(os0), _flatZinc(flatZinc) {}
 
     void p(const Type& type, const Expression* e) {
       switch (type.ti()) {
@@ -206,12 +229,7 @@ namespace MiniZinc {
         break;
       case Expression::E_FLOATLIT:
         {
-          std::ostringstream oss;
-          oss << std::setprecision(std::numeric_limits<double>::digits10+1);
-          oss << e->cast<FloatLit>()->v();
-          if (oss.str().find("e") == std::string::npos && oss.str().find(".") == std::string::npos)
-            oss << ".0";
-          os << oss.str();
+          ppFloatVal(os, e->cast<FloatLit>()->v());
         }
         break;
       case Expression::E_SETLIT:
@@ -242,6 +260,43 @@ namespace MiniZinc {
               if (!sl.isv()->max(sl.isv()->size()-1).isFinite())
                 os << "++" << sl.isv()->min(sl.isv()->size()-1) << ".." << sl.isv()->max(sl.isv()->size()-1);
             }
+          } else if (sl.fsv()) {
+            if (sl.fsv()->size()==0) {
+              os << (_flatZinc ? "1.0..0.0" : "{}");
+            } else if (sl.fsv()->size()==1) {
+              ppFloatVal(os, sl.fsv()->min(0));
+              os << "..";
+              ppFloatVal(os, sl.fsv()->max(0));
+            } else {
+              bool allSingleton = true;
+              for (FloatSetRanges isr(sl.fsv()); isr(); ++isr) {
+                if (isr.min() != isr.max()) {
+                  allSingleton = false;
+                  break;
+                }
+              }
+              if (allSingleton) {
+                os << "{";
+                bool first = true;
+                for (FloatSetRanges isr(sl.fsv()); isr(); ++isr) {
+                  if (!first)
+                    os << ",";
+                  first = false;
+                  ppFloatVal(os, isr.min());
+                }
+                os << "}";
+              } else {
+                bool first = true;
+                for (FloatSetRanges isr(sl.fsv()); isr(); ++isr) {
+                  if (!first)
+                    os << "++";
+                  first = false;
+                  ppFloatVal(os, isr.min());
+                  os << "..";
+                  ppFloatVal(os, isr.max());
+                }
+              }
+            }
           } else {
             os << "{";
             for (unsigned int i = 0; i < sl.v().size(); i++) {
@@ -265,10 +320,12 @@ namespace MiniZinc {
             os << "<>";
           } else {
             const Id* id = e->cast<Id>();
+            if(id->decl())
+              id = id->decl()->id();
             if (id->idn() == -1) {
               os << id->v();
             } else {
-              os << "X_INTRODUCED_" << id->idn();
+              os << "X_INTRODUCED_" << id->idn() << "_";
             }
           }
         }
@@ -285,17 +342,17 @@ namespace MiniZinc {
           int n = al.dims();
           if (n == 1 && al.min(0) == 1) {
             os << "[";
-            for (unsigned int i = 0; i < al.v().size(); i++) {
-              p(al.v()[i]);
-              if (i<al.v().size()-1)
+            for (unsigned int i = 0; i < al.size(); i++) {
+              p(al[i]);
+              if (i<al.size()-1)
                 os << ",";
             }
             os << "]";
-          } else if (n == 2 && al.min(0) == 1 && al.min(1) == 1) {
+          } else if (n == 2 && al.min(0) == 1 && al.min(1) == 1 && al.max(1) != 0) {
             os << "[|";
             for (int i = 0; i < al.max(0); i++) {
               for (int j = 0; j < al.max(1); j++) {
-                p(al.v()[i * al.max(1) + j]);
+                p(al[i * al.max(1) + j]);
                 if (j < al.max(1)-1)
                   os << ",";
               }
@@ -310,9 +367,9 @@ namespace MiniZinc {
               os << ",";
             }
             os << "[";
-            for (unsigned int i = 0; i < al.v().size(); i++) {
-              p(al.v()[i]);
-              if (i<al.v().size()-1)
+            for (unsigned int i = 0; i < al.size(); i++) {
+              p(al[i]);
+              if (i<al.size()-1)
                 os << ",";
             }
             os << "])";
@@ -344,15 +401,20 @@ namespace MiniZinc {
               if (j < c.n_decls(i)-1)
                 os << ",";
             }
-            os << " in ";
-            p(c.in(i));
+            if (c.in(i)==NULL) {
+              os << " = ";
+              p(c.where(i));
+            } else {
+              os << " in ";
+              p(c.in(i));
+              if (c.where(i) != NULL) {
+                os << " where ";
+                p(c.where(i));
+              }
+            }
             if (i < c.n_generators())
               os << ", ";
           }
-          if (c.where() != NULL) {
-            os << " where ";
-            p(c.where());
-          }
           os << (c.set() ? "}" : "]");
         }
         break;
@@ -389,6 +451,9 @@ namespace MiniZinc {
           case BOT_MULT:
             os<<"*";
             break;
+          case BOT_POW:
+            os<<"^";
+            break;
           case BOT_DIV:
             os<<"/";
             break;
@@ -502,9 +567,9 @@ namespace MiniZinc {
         {
           const Call& c = *e->cast<Call>();
           os << c.id() << "(";
-          for (unsigned int i = 0; i < c.args().size(); i++) {
-            p(c.args()[i]);
-            if (i < c.args().size()-1)
+          for (unsigned int i = 0; i < c.n_args(); i++) {
+            p(c.arg(i));
+            if (i < c.n_args()-1)
               os << ",";
           }
           os << ")";
@@ -514,10 +579,13 @@ namespace MiniZinc {
         {
           const VarDecl& vd = *e->cast<VarDecl>();
           p(vd.ti());
+          if (!vd.ti()->isEnum()) {
+            os << ":";
+          }
           if (vd.id()->idn() != -1) {
-            os << ": X_INTRODUCED_" << vd.id()->idn();
+            os << " X_INTRODUCED_" << vd.id()->idn() << "_";
           } else if (vd.id()->v().size() != 0)
-            os << ": " << vd.id()->v();
+            os << " " << vd.id()->v();
           if (vd.introduced()) {
             os << " ::var_is_introduced ";
           }
@@ -549,16 +617,22 @@ namespace MiniZinc {
       case Expression::E_TI:
         {
           const TypeInst& ti = *e->cast<TypeInst>();
-          if (ti.isarray()) {
-            os << "array [";
-            for (unsigned int i = 0; i < ti.ranges().size(); i++) {
-              p(Type::parint(), ti.ranges()[i]);
-              if (i < ti.ranges().size()-1)
-                os << ",";
+          if (ti.isEnum()) {
+            os << "enum";
+          } else if (env) {
+            os << ti.type().toString(*env);
+          } else {
+            if (ti.isarray()) {
+              os << "array [";
+              for (unsigned int i = 0; i < ti.ranges().size(); i++) {
+                p(Type::parint(), ti.ranges()[i]);
+                if (i < ti.ranges().size()-1)
+                  os << ",";
+              }
+              os << "] of ";
             }
-            os << "] of ";
+            p(ti.type(),ti.domain());
           }
-          p(ti.type(),ti.domain());
         }
       }
       if (!e->isa<VarDecl>()) {
@@ -888,7 +962,7 @@ namespace MiniZinc {
   }
   void
   Line::addString(const std::string& s) {
-    lineLength += s.size();
+    lineLength += static_cast<int>(s.size());
     text.push_back(s);
   }
   void
@@ -1045,10 +1119,7 @@ namespace MiniZinc {
     }
     ret mapFloatLit(const FloatLit& fl) {
       std::ostringstream oss;
-      oss << std::setprecision(std::numeric_limits<double>::digits10+1);
-      oss << fl.v();
-      if (oss.str().find("e") == std::string::npos && oss.str().find(".") == std::string::npos)
-        oss << ".0";
+      ppFloatVal(oss, fl.v());
       return new StringDocument(oss.str());
     }
     ret mapSetLit(const SetLit& sl) {
@@ -1077,6 +1148,33 @@ namespace MiniZinc {
             dl->addDocumentToList(new StringDocument(oss.str()));
           }
         }
+      } else if (sl.fsv()) {
+        if (sl.fsv()->size()==0) {
+          dl = new DocumentList("1.0..0.0","","");
+        } else if (sl.fsv()->size()==1) {
+          dl = new DocumentList("", "..", "");
+          {
+            std::ostringstream oss;
+            ppFloatVal(oss, sl.fsv()->min(0));
+            dl->addDocumentToList(new StringDocument(oss.str()));
+          }
+          {
+            std::ostringstream oss;
+            ppFloatVal(oss, sl.fsv()->max(0));
+            dl->addDocumentToList(new StringDocument(oss.str()));
+          }
+        } else {
+          dl = new DocumentList("", "++", "", true);
+          FloatSetRanges fsr(sl.fsv());
+          for (; fsr(); ++fsr) {
+            std::ostringstream oss;
+            ppFloatVal(oss, fsr.min());
+            oss << "..";
+            ppFloatVal(oss, fsr.max());
+            dl->addDocumentToList(new StringDocument(oss.str()));
+          }
+        }
+
       } else {
         dl = new DocumentList("{", ", ", "}", true);
         for (unsigned int i = 0; i < sl.v().size(); i++) {
@@ -1102,7 +1200,7 @@ namespace MiniZinc {
         return new StringDocument(id.v().str());
       else {
         std::ostringstream oss;
-        oss << "X_INTRODUCED_" << id.idn();
+        oss << "X_INTRODUCED_" << id.idn() << "_";
         return new StringDocument(oss.str());
       }
     }
@@ -1118,15 +1216,15 @@ namespace MiniZinc {
       int n = al.dims();
       if (n == 1 && al.min(0) == 1) {
         dl = new DocumentList("[", ", ", "]");
-        for (unsigned int i = 0; i < al.v().size(); i++)
-          dl->addDocumentToList(expressionToDocument(al.v()[i]));
+        for (unsigned int i = 0; i < al.size(); i++)
+          dl->addDocumentToList(expressionToDocument(al[i]));
       } else if (n == 2 && al.min(0) == 1 && al.min(1) == 1) {
         dl = new DocumentList("[| ", " | ", " |]");
         for (int i = 0; i < al.max(0); i++) {
           DocumentList* row = new DocumentList("", ", ", "");
           for (int j = 0; j < al.max(1); j++) {
             row->
-              addDocumentToList(expressionToDocument(al.v()[i * al.max(1) + j]));
+              addDocumentToList(expressionToDocument(al[i * al.max(1) + j]));
           }
           dl->addDocumentToList(row);
           if (i != al.max(0) - 1)
@@ -1145,8 +1243,8 @@ namespace MiniZinc {
           args->addStringToList(oss.str());
         }
         DocumentList* array = new DocumentList("[", ", ", "]");
-        for (unsigned int i = 0; i < al.v().size(); i++)
-          array->addDocumentToList(expressionToDocument(al.v()[i]));
+        for (unsigned int i = 0; i < al.size(); i++)
+          array->addDocumentToList(expressionToDocument(al[i]));
         args->addDocumentToList(array);
         dl->addDocumentToList(args);
       }
@@ -1180,15 +1278,20 @@ namespace MiniZinc {
           idents->addStringToList(c.decl(i, j)->id()->v().str());
         }
         gen->addDocumentToList(idents);
-        gen->addStringToList(" in ");
-        gen->addDocumentToList(expressionToDocument(c.in(i)));
+        if (c.in(i)==NULL) {
+          gen->addStringToList(" = ");
+          gen->addDocumentToList(expressionToDocument(c.where(i)));
+        } else {
+          gen->addStringToList(" in ");
+          gen->addDocumentToList(expressionToDocument(c.in(i)));
+          if (c.where(i) != NULL) {
+            gen->addStringToList(" where ");
+            gen->addDocumentToList(expressionToDocument(c.where(i)));
+          }
+        }
         generators->addDocumentToList(gen);
       }
       head->addDocumentToList(generators);
-      if (c.where() != NULL) {
-        head->addStringToList("where");
-        head->addDocumentToList(expressionToDocument(c.where()));
-      }
       dl->addDocumentToList(head);
 
       return dl;
@@ -1243,6 +1346,9 @@ namespace MiniZinc {
       case BOT_MULT:
         op = "*";
         break;
+      case BOT_POW:
+        op = "^";
+        break;
       case BOT_DIV:
         op = "/";
         break;
@@ -1366,7 +1472,7 @@ namespace MiniZinc {
       return dl;
     }
     ret mapCall(const Call& c) {
-      if (c.args().size() == 1) {
+      if (c.n_args() == 1) {
         /*
          * if we have only one argument, and this is an array comprehension,
          * we convert it into the following syntax
@@ -1375,7 +1481,7 @@ namespace MiniZinc {
          * forall (i in 1..10) (f(i,j))
          */
 
-        const Expression* e = c.args()[0];
+        const Expression* e = c.arg(0);
         if (e->isa<Comprehension>()) {
           const Comprehension* com = e->cast<Comprehension>();
           if (!com->set()) {
@@ -1392,17 +1498,22 @@ namespace MiniZinc {
                   com->decl(i,j)->id()->v().str());
               }
               gen->addDocumentToList(idents);
-              gen->addStringToList(" in ");
-              gen->addDocumentToList(expressionToDocument(com->in(i)));
+              if (com->in(i) == NULL) {
+                gen->addStringToList(" = ");
+                gen->addDocumentToList(expressionToDocument(com->where(i)));
+              } else {
+                gen->addStringToList(" in ");
+                gen->addDocumentToList(expressionToDocument(com->in(i)));
+                if (com->where(i) != NULL) {
+                  gen->addStringToList(" where ");
+                  gen->addDocumentToList(expressionToDocument(com->where(i)));
+                }
+              }
               generators->addDocumentToList(gen);
             }
 
             args->addStringToList("(");
             args->addDocumentToList(generators);
-            if (com->where() != NULL) {
-              args->addStringToList("where");
-              args->addDocumentToList(expressionToDocument(com->where()));
-            }
             args->addStringToList(")");
 
             args->addStringToList("(");
@@ -1420,8 +1531,8 @@ namespace MiniZinc {
       }
       std::string beg = c.id().str() + "(";
       DocumentList* dl = new DocumentList(beg, ", ", ")");
-      for (unsigned int i = 0; i < c.args().size(); i++) {
-        dl->addDocumentToList(expressionToDocument(c.args()[i]));
+      for (unsigned int i = 0; i < c.n_args(); i++) {
+        dl->addDocumentToList(expressionToDocument(c.arg(i)));
       }
       return dl;
 
@@ -1436,7 +1547,7 @@ namespace MiniZinc {
         dl->addStringToList(vd.id()->v().str());
       } else {
         std::ostringstream oss;
-        oss << "X_INTRODUCED_" << vd.id()->idn();
+        oss << "X_INTRODUCED_" << vd.id()->idn() << "_";
         dl->addStringToList(oss.str());
       }
         
@@ -1702,7 +1813,7 @@ namespace MiniZinc {
   void
   PrettyPrinter::print(std::ostream& os) const {
     std::vector<Line>::const_iterator it;
-    int nItems = items.size();
+    int nItems = static_cast<int>(items.size());
     for (int item = 0; item < nItems; item++) {
       for (it = items[item].begin(); it != items[item].end(); it++) {
         it->print(os);
@@ -1750,7 +1861,7 @@ namespace MiniZinc {
   void PrettyPrinter::printString(const std::string& s, bool alignment,
                                   int alignmentCol) {
     Line& l = items[currentItem][currentLine];
-    int size = s.size();
+    int size = static_cast<int>(s.size());
     if (size <= l.getSpaceLeft(maxwidth)) {
       l.addString(s);
     } else {
@@ -1776,8 +1887,8 @@ namespace MiniZinc {
     int currentCol = items[currentItem][currentLine].getIndentation()
         + items[currentItem][currentLine].getLength();
     int newAlignmentCol =
-        _alignment ? currentCol + beginToken.size() : alignmentCol;
-    int vectorSize = ld.size();
+        _alignment ? currentCol + static_cast<int>(beginToken.size()) : alignmentCol;
+    int vectorSize = static_cast<int>(ld.size());
     int lastVisibleElementIndex;
     for (int i = 0; i < vectorSize; i++) {
       if (!dynamic_cast<BreakPoint*>(ld[i]))
@@ -1848,8 +1959,8 @@ namespace MiniZinc {
     return true;
   }
 
-  Printer::Printer(std::ostream& os, int width, bool flatZinc)
-  : ism(NULL), printer(NULL), _os(os), _width(width), _flatZinc(flatZinc) {}
+  Printer::Printer(std::ostream& os, int width, bool flatZinc, EnvI* env0)
+  : env(env0), ism(NULL), printer(NULL), _os(os), _width(width), _flatZinc(flatZinc) {}
   void
   Printer::init(void) {
     if (ism==NULL) {
@@ -1902,7 +2013,7 @@ namespace MiniZinc {
   void
   Printer::print(const Expression* e) {
     if (_width==0) {
-      PlainPrinter p(_os,_flatZinc); p.p(e);
+      PlainPrinter p(_os,_flatZinc,env); p.p(e);
     } else {
       init();
       Document* d = expressionToDocument(e);
@@ -1913,7 +2024,7 @@ namespace MiniZinc {
   void
   Printer::print(const Item* i) {
     if (_width==0) {
-      PlainPrinter p(_os,_flatZinc); p.p(i);
+      PlainPrinter p(_os,_flatZinc,env); p.p(i);
     } else {
       init();
       p(i);
@@ -1922,7 +2033,7 @@ namespace MiniZinc {
   void
   Printer::print(const Model* m) {
     if (_width==0) {
-      PlainPrinter p(_os,_flatZinc);
+      PlainPrinter p(_os,_flatZinc,env);
       for (unsigned int i = 0; i < m->size(); i++) {
         p.p((*m)[i]);
       }
